from __future__ import print_function
"""
Elasticsearch Extension.
"""
import json
import requests
import logging
from tubing import sinks, tubes

logger = logging.getLogger('tubing.ext.elasticsearch')


class DocUpdate(object):  # pragma: no cover
    """
    DocUpdate is an ElasticSearch document update object. It is meant to be
    used with BulkBatcher and returns an action and update.
    """

    def __init__(
        self,
        esid,
        doc,
        doc_type,
        parent_esid=None,
        doc_as_upsert=True
    ):
        """
        esid is an elasticsearch _id
        parent_esid is the elasticsearch _id of parent if parent exists
        doc is a dict
        """
        self.doc = doc
        self.esid = esid
        self.parent_esid = parent_esid
        self.doc_as_upsert = doc_as_upsert
        self.doc_type = doc_type

    def action(self, encoding):
        action = dict(update=dict(_id=self.esid, _type=self.doc_type,),)
        if self.parent_esid:
            action['update']['parent'] = self.parent_esid
        return json.dumps(action).encode(encoding)

    def update(self, encoding):
        return json.dumps(
            dict(
                doc_as_upsert=self.doc_as_upsert,
                doc=self.doc,
            )
        ).encode(encoding)

    def serialize(self, encoding):
        return self.action(encoding) + b"\n" + self.update(encoding) + b"\n"


class ElasticSearchError(Exception):
    """
    ElasticSearchError message is the text response from the elasticsearch
    server.
    """
    pass


class BulkUpdateTransformer(object):
    """
    BulkSinkWriter writes bulk updates to Elastic Search. If username or
    password is None, then auth will be skipped.
    """

    def __init__(self, encoding='utf-8'):
        self.encoding = encoding

    def transform(self, chunk):
        data = []
        for update in chunk:
            data.append(update.serialize(self.encoding))
        return data


PrepareBulkUpdate = tubes.MakeTransformerTubeFactory(
    BulkUpdateTransformer,
    default_chunk_size=2**10
)


def BulkUpdate(
    base_url,
    index,
    username=None,
    password=None,
    chunk_size=50,
    chunks_per_post=20,
    fail_on_error=True,
):
    """
    Docs per post is chunk_size * chunks_per_post.
    """
    url = "%s/%s/_bulk" % (base_url, index)

    def response_handler(resp):
        try:
            try:
                resp_obj = json.loads(resp.text)
            except ValueError:
                raise ElasticSearchError("invalid response: '%s'" % (resp.text))

            if resp_obj['errors']:
                raise ElasticSearchError(
                    "errors in response: '%s'" % (resp.text)
                )
        except:
            logger.exception("%s %s" % (resp, resp.text))
            if fail_on_error:
                raise

    return sinks.HTTPPost(
        url=url,
        username=username,
        password=password,
        chunk_size=chunk_size,
        chunks_per_post=chunks_per_post,
        response_handler=response_handler,
    )



from __future__ import print_function
"""
Tubing sources are defined here. If you want to make your own Source, create a
reader class with a read(amt) function, where amt is the amount of `stuff` to
read. MakeSourceFactory can generate a Source from your Reader. Ex::

    class MyReader(object):
        def read(self, amt):
            return [0] * amt, False

    MySource = MakeSourceFactory(MyReader)

The read(amt) function should return a chunk of data and a boolean indicated if
we've reached EOF. Unlike normal python streams, it's ok to return empty sets
and it won't close the stream. Only returning True as the second parameter
indicates that the stream is closed.
"""
import logging
import socket
import signal
import io
from requests import Session, Request
from tubing import compat

logger = logging.getLogger('tubing.sources')

HANDLERS = []

def handle_signals(*signals):
    def handle(*args, **kwargs):
        for handle in HANDLERS:
            handle()
    for sig in signals:
        signal.signal(sig, handle)

handle_signals(signal.SIGTERM, signal.SIGINT, signal.SIGHUP)

class MakeSourceFactory(object):
    """
    MakeSourceFactory takes a reader object and returns a Source factory.
    """

    def __init__(self, reader_cls):
        self.reader_cls = reader_cls

    def __call__(self, *args, **kwargs):
        reader = self.reader_cls(*args, **kwargs)
        src = Source(reader)
        if hasattr(reader, 'interrupt'):
            HANDLERS.append(reader.interrupt)
        return src


@compat.python_2_unicode_compatible
class Source(object):
    """
    Source is a wrapper for Readers that allows piping.
    """

    def __init__(self, source):
        self.source = source

    def read(self, amt):
        return self.source.read(amt)

    def __or__(self, other):
        return self.pipe(other)

    def pipe(self, other):
        return other.receive(self)

    def __str__(self):
        return "<tubing.sources.Source(%s)>" % (self.source)


class ObjectReader(object):
    """
    ObjectReader outputs a list of objects.
    """

    def __init__(self, objs):
        self.objs = objs

    def read(self, amt=None):
        r, self.objs = self.objs[:amt], self.objs[amt or len(self.objs):]
        return r, len(self.objs) == 0


Objects = MakeSourceFactory(ObjectReader)


class FileReader(object):
    """
    FileReader outputs bytes.
    """

    def __init__(self, filename, mode="rb"):
        self.filename = filename
        self.f = open(self.filename, mode)

    def read(self, amt=None):
        chunk = self.f.read(amt)
        if chunk:
            return chunk, False
        else:
            return '', True

    def __unicode__(self):
        return u"<tubing.sources.File %s>" % (self.filename)

    def __str__(self):
        return unicode(self).encode('utf-8')


File = MakeSourceFactory(FileReader)


class SocketReader(object):
    """
    UDPReader binds and reads from a UDP socket.
    """
    def __init__(self, ip, port, *args):
        self.sock = socket.socket(*args)
        self.sock.bind((ip, port))
        self.sock.settimeout(0.1)
        self.eof = False

    def read(self, amt=None):
        if self.eof:
            return b'', True

        try:
            data, addr = self.sock.recvfrom(amt)
            logger.debug("%s >> %s" % (addr, data))
            return data, False
        except socket.timeout:
            return b'', False
        except socket.error:
            return b'', True

    def interrupt(self):
        self.eof = True


Socket = MakeSourceFactory(SocketReader)


class BytesReader(object):
    def __init__(self, byts):
        self.io = io.BytesIO(byts)

    def read(self, amt=None):
        r = self.io.read(amt)
        if r:
            return r, False
        else:
            return b'', True

Bytes = MakeSourceFactory(BytesReader)


class IOReader(object):
    def __init__(self, stream):
        self.io = stream

    def read(self, amt=None):
        r = self.io.read(amt)
        if r:
            return r, False
        else:
            return b'', True

IO = MakeSourceFactory(IOReader)


class HTTPReader(object):
    def __init__(self, method, url, *args, **kwargs):
        s = Session()
        s.stream = True
        r = Request(method, url, *args, **kwargs)
        self.stream = s.send(r.prepare()).raw

    def read(self, amt=None):
        r = self.stream.read(amt)
        if r:
            return r, False
        else:
            return b'', True

HTTP = MakeSourceFactory(HTTPReader)



from __future__ import print_function
"""
Tubing sinks are targets for streams of data. To make your own sink, define a Writer with a write(chunk) and optional close() and abort() functions and pass it to MakeSink. if present, close() will be called on the upstream EOF and abort() will be called on any exception in the pipeline. Here's an example::

    class MyWriter(object):
        def write(self, chunk):
            print chunk

    MySink = MakeSink(MyWriter)
"""
import requests
import logging
import io
import functools

logger = logging.getLogger('tubing.sinks')


class SinkRunner(object):

    def __init__(self, source, sink, chunk_size):
        self.source = source
        self.sink = sink
        self.chunk_size = chunk_size

    def __call__(self):
        try:
            logger.debug("reading %s" % (self.source))
            chunk, eof = self.source.read(self.chunk_size)
            self.sink.write(chunk)
            while not eof:
                chunk, eof = self.source.read(self.chunk_size)
                self.sink.write(chunk)
            hasattr(self.sink, 'close') and self.sink.close()
            return self.sink.writer
        except:
            logger.exception("Pipe failed")
            hasattr(self.sink, 'abort') and self.sink.abort()
            raise


class Sink(object):

    def __init__(self, writer, chunk_size=2**16):
        self.writer = writer
        self.chunk_size = chunk_size

    def write(self, chunk):
        self.writer.write(chunk)
        return self

    def close(self):
        hasattr(self.writer, 'close') and self.writer.close()

    def abort(self):
        hasattr(self.writer, 'abort') and self.writer.abort()

    def receive(self, source):
        SinkRunner(source, self, self.chunk_size)()
        return self.writer


def SinkFactory(writer_cls, default_chunk_size=2**16, *args, **kwargs):
    if kwargs.get('chunk_size'):
        chunk_size = kwargs['chunk_size']
        del kwargs['chunk_size']
    else:
        chunk_size = default_chunk_size

    writer = writer_cls(*args, **kwargs)
    return Sink(writer, chunk_size)


def MakeSinkFactory(sink_cls, default_chunk_size=2**16):
    return functools.partial(SinkFactory, sink_cls, default_chunk_size)


class ObjectsSink(list):
    """
    Writes an object stream to a list.
    """

    def write(self, objs):
        self.extend(objs)


Objects = MakeSinkFactory(ObjectsSink, 2**4)


class BytesWriter(io.BytesIO):
    """
    BytesWriter collects a stream of bytes and saves it to the result member
    on close().
    """

    def __init__(self, *args, **kwargs):
        self.result = None
        super(BytesWriter, self).__init__(*args, **kwargs)

    def close(self):
        self.result = self.getvalue()
        super(BytesWriter, self).close()

    def abort(self):
        super(BytesWriter, self).close()


Bytes = MakeSinkFactory(BytesWriter, 2**16)


class FileWriter(object):

    def __init__(self, *args, **kwargs):
        self.f = open(*args, **kwargs)

    def write(self, chunk):
        self.f.write(chunk)

    def close(self):
        self.f.close()

    def abort(self):
        # TODO: probably delete the file
        self.f.close()


File = MakeSinkFactory(FileWriter)


class HTTPPost(object):
    """
    HTTPPost doesn't support the write method, and therefore can not be used
    with tubes.Tee.
    """
    def __init__(self, url, username=None, password=None, chunk_size=2**32,
                 chunks_per_post=2**10, response_handler=lambda _: None):
        self.url = url
        if username:
            self.auth = (username, password)
        else:
            self.auth = None
        self.chunk_size = chunk_size
        self.chunks_per_post = chunks_per_post
        self.response_handler = response_handler
        self.eof = False

    def gen(self):
        """
        We have to do this goofy shit because requests.post doesn't give
        access to the socket directly. In order to stream, we need to pass
        a generator object to requests.
        """
        for _ in range(self.chunks_per_post):
            r, self.eof = self.source.read(self.chunk_size)
            yield r
            if self.eof:
                return

    def receive(self, other):
        self.source = other
        while not self.eof:
            r = requests.post(self.url, data=self.gen(), auth=self.auth)
            self.response_handler(r)


class DebugPrinter(object):

    def write(self, chunk):
        logger.debug(chunk)

    def close(self):
        logger.debug("CLOSED")

    def abort(self):
        logger.debug("ABORTED")


Debugger = MakeSinkFactory(DebugPrinter)



"""
There are two types of tubing users. Irish users, and users who wish they were
Irish.  j/k. Really, they are:

 - plebe users want to use whatever we have in tubing.
 - noble users who want to extend tubing to their own devices.
 - royal users want to contribute to tubing.

We are socialists, so we want to take care of the plebes, first and foremost.
They are also the easiest to satisfy. For them, we have tubes. Tubes are easy
to use and understand.

Nobles are very important too, but harder to satisfy. We never know what crazy
plans they'll have in mind, so we must be ready. They need the tools to build
new tubes that extend our apparatus in unexpected ways.

For the benefit of the nobles, and ourselves, we're going to outline exactly
how things work now. This documentation is also an exercise in simplifying the
code base.

We'll call a tubing pipeline an apparatus. An apparatus has a Source, zero to
many Tubes, and a Sink.

A stream is what we call the units flowing through our Tubes.

Tubes
=====

The easiest way to extend tubing is to create a Transformer, and use MakeTransformerTubeFactory
to turn it into a Tube. A Transformer has the following interface::

    class Transformer(object):
        def transform(self, chunk):
            return new_chunk

    NewTube = MakeTransformerTubeFactory(Transformer)

A chunk is an iterable of whatever type of stream we are working on, whether it
be bytes, Unicode characters, strings or python objects.  We can index it,
slice it, or iterate over it. `transform` simple takes a chunk, and makes a new
chunk out of it. `MakeTransformerTubeFactory` will take care of all the dirty
work. Transformers are enough for most tasks, but if you need to do something
more complex, you may need to go deeper.

.. image:: http://i.imgur.com/DyPouyL.png
    alt: Leonardo DiCaprio

First let's describe how tubes work in more detail. Here's the Tube interface::

    class TubeFactory(object):
        # This is what we export, and what is called when users create a tube.
        # The syntax looks like this:
        # SourceFactory() | [TubeFactory()...] | SinkFactory()
        def __call__(self, *args, **kwargs):
            return Tube()

    # - or -

    def TubeFactory(*args, **kwargs):
        return Tube()

    class Tube(object):
        def receive(self, source):
            # return a TubeWorker
            tw = TubeWorker
            tw.source = source
            return tw

    class TubeWorker(object):
        def tube(self, receiver):
            # receiver is they guy who will call our `read` method. Either
            # another Tube or a Sink.
            return receiver.receive(self)

        def __or__(self, *args, **kwargs):
            # Our reason for existing.
            return self.tube(*args, **kwargs)

        def read(self, amt=None):
            # our receiver will call this guy. We return a tuple here of
            # `chunk, eof`.  We should return a chunk of len amt of whatever
            # type of object we produce. If we've exhausted our upstream
            # source, then we should return True as the second element of our
            # tuple.
            return [], True

A TubeFactory is what plebes deal with. As you can see, it can be an object or
a function, depending on your style. It's easier for me to reason about state
with an object, but if you prefer a closure, go for it! Classes are just
closures with more verbose states.

When a plebe is setting up some tubing, the TubeFactory returns a Tube, but this
isn't the last object we'll create. The Tube doesn't have a source connected, so
it's sort of useless. It's just a placeholder waiting for a source. As soon as
it gets a source, it will hand off all of it's duties to a TubeWorker.

A TubeWorker is ready to read from it's source, but it doesn't. TubeWorkers are
pretty lazy and need someone else to tell them what to do. That's where a receiver
comes in hand. A receiver can be another Tube, or a Sink.  If it's another Tube,
you know the drill. It's just another lazy guy that will only tell his source to
read when his boss tells him to read. Ultimately, the only guy who wants to do
any work is the Sink. At the end of the chain, a sink's receive function will
be called, and he'll get everyone to work.

Technically, we could split the TubeWorker interface into two parts, but it's not
really necessary since they share the same state. We could also combine TubeFactory,
Tube and TubeWorker, and just build up state overtime. I've seriously consider this,
but I don't know, this feels better. I admit, it is a little complicated, but one advantage you get is that you can do something like this::

    from tubing import tubes

    tube = tubes.GZip(chunk_size=2**32)

    source | tube | output1
    source | tube | output2

Since tube is a factory and not an object, each tubeline will have it's own
state. tubeline.... I just made that up. That's an execution pipeline in
tubing. But we don't want to call it a pipeline, it's a tubeline. Maybe there's
a better name? I picture a chemistry lab with a bunch of transparent tubes
connected to beakers and things like that.

.. image:: http://imgur.com/jTtHITH.jpg
    :alt: chemistry lab

Let's call it an apparatus.

MakeTransformerTubeFactory
--------------------------

So how does MakeTransformerTubeFactory turn a Transformer into a
TubeFactory?  MakeTransformerTubeFactory is a utility that creates a
function that wrap a transformer in a tube. Sort of complicate, eh? I'm sorry
about that, but let's see if we can break it down.

MakeTransformerTubeFactory returns a partial function out of the TransformerTube
instantiation. For the uninitiated, a partial is just a new version
of a function with some of the parameters already filled in. So we're currying
the transformer_cls and the default_chunk_size back to the plebes. They
can fill in the rest of the details and get back a TransformerTube.

The TransformerTubeWorker is where most of the hard work happens. There's a
bunch of code related to reading just enough chunks from our source to satisfy
our receiver. Remember, Workers are lazy, that's good.

default_chunk_size is sort of important, by default it's something like 32k. It's
the size of the chunks that we request from upstream, in the read function (amt).
That's great for byte streams(maybe?), but it's not that great for large objects.
You'll probably want to set it if you are using something other than bytes. It
can be overridden by plebes, this is just the default if they don't specify it.
Remember, we should be making the plebes job easy, so try and be a nice noble
and set it to something sensible.

We've explained Tubes, very well I might add. And it's a good thing. They are
the most complicated bit in tubing. All that's left is Sources and Sinks.

Sources
=======

TODO

Sinks
=====

TODO

Things You Can't do with Tubing
===============================

 - Tee to another apparatus
 - async programming
 - your laundry

"""
import pkgutil

try:
    version = pkgutil.get_data(__name__, 'VERSION').decode('utf-8')
except IOError:  # pragma: no cover
    version = '9999'

__version__ = version



from __future__ import print_function
"""
S3 Tubing extension.
"""

import boto3
import logging
from tubing import sources, sinks, compat

logger = logging.getLogger('tubing.ext.s3')


@compat.python_2_unicode_compatible
class S3Reader(object):  # pragma: no cover
    """
    Read file from S3. Expects AWS environmental variables to be set.
    """

    def __init__(self, bucket, key):  # pragma: no cover
        """
        Create an S3 Source stream.
        """
        s3 = boto3.client('s3')
        self.bucket = bucket
        self.key = key
        self.response = s3.get_object(Bucket=bucket, Key=key)

    def read(self, amt):
        r = self.response['Body'].read(amt)
        return r or b'', not r

    def __str__(self):
        return "<tubing.ext.s3.S3Source s3://%s/%s>" % (self.bucket, self.key)


S3Source = sources.MakeSourceFactory(S3Reader)


class MultipartWriter(object):  # pragma: no cover
    """
    Send file to S3. Expects AWS environmental variables to be set.
    """

    def __init__(self, bucket, key):
        """
        Initiate the multi-part upload.
        """
        self.s3 = boto3.client('s3')
        self.bucket = bucket
        self.key = key
        logger.debug("Creating upload for %s %s" % (bucket, key))
        upload = self.s3.create_multipart_upload(Bucket=bucket, Key=key)
        self.upload_id = upload["UploadId"]
        # start with 1
        self.part_number = 1
        self.part_info = dict(Parts=[])

    def track_part(self, part_number, part):
        """
        Keep track of all parts so we can finalize multipart upload.
        """
        self.part_info['Parts'].append(
            dict(
                PartNumber=part_number,
                ETag=part['ETag'],
            )
        )

    def write(self, chunk):
        """
        Upload a part.
        """
        if len(chunk):
            logger.debug("Posting %s [%d]" % (self.part_number, len(chunk)))
            part = self.s3.upload_part(
                Bucket=self.bucket,
                Key=self.key,
                PartNumber=self.part_number,
                UploadId=self.upload_id,
                Body=chunk,
            )
            self.track_part(self.part_number, part)
            self.part_number += 1

    def close(self):
        """
        Finalize upload.
        """
        if self.part_number > 1:
            self.s3.complete_multipart_upload(
                Bucket=self.bucket,
                Key=self.key,
                UploadId=self.upload_id,
                MultipartUpload=self.part_info,
            )
        else:
            self.abort()

    def abort(self):
        """
        Something failed, abort!
        """
        self.s3.abort_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
        )


MultipartUpload = sinks.MakeSinkFactory(MultipartWriter, default_chunk_size=2**33)



"""
compat provides tools to make code compatible across python versions.
"""
import sys

PY2 = sys.version_info[0] == 2


def python_2_unicode_compatible(klass):
    """
    *lifted from Django*
    A decorator that defines __unicode__ and __str__ methods under Python 2.
    Under Python 3 it does nothing.

    To support Python 2 and 3 with a single code base, define a __str__ method
    returning text and apply this decorator to the class.
    """
    if PY2:
        if '__str__' not in klass.__dict__:
            raise ValueError(
                "@python_2_unicode_compatible cannot be applied "
                "to %s because it doesn't define __str__()." % klass.__name__
            )
        klass.__unicode__ = klass.__str__
        klass.__str__ = lambda self: self.__unicode__().encode('utf-8')
    return klass
    if PY2:
        if '__str__' not in klass.__dict__:
            raise ValueError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import os.path
import sys

from codecs import open
from datetime import datetime
from setuptools import setup, find_packages

minor_version = "0.0.2"


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()


VERSIONFILE = os.path.join(os.path.dirname(__file__), 'tubing/VERSION')
if os.path.exists(os.path.join(os.path.dirname(__file__), 'tubing/VERSION')):
    with open(VERSIONFILE, "r", "utf-8") as f:
        version = f.read().strip()
else:
    if os.environ.get("TRAVIS_BRANCH") == "release":
        revision = "r" + os.environ.get("TRAVIS_BUILD_NUMBER")
        version = minor_version + revision
    else:
        date = datetime.now().strftime(".dev%Y%m%d%H%M%S")
        revision = os.environ.get("REVISION", date)
        version = minor_version + revision

with open(VERSIONFILE, "w", "utf-8") as f:
    f.write(version)

with open("README.rst", "r", "utf-8") as f:
    readme = f.read()


setup(
    name="tubing",
    version=version,
    description="Python I/O pipe utilities",
    long_description=readme,
    author="Bob Corsaro",
    author_email="rcorsaro@gmail.com",
    url="http://github.com/dokipen/tubing",
    package_data={"": ["LICENSE"], "tubing": ["VERSION"]},
    packages=find_packages(exclude='tests'),
    include_package_data=True,
    install_requires=[
        "requests",
        "urllib3",
    ],
    license="MIT",
    zip_safe=False,
    classifiers=(
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ),
    extras_require={
        "build": [
            "yapf",
            "pyflakes",
            "nose",
            "unittest2",
            "coveralls",
            "boto3",
        ],
        "s3": [
            "boto3",
        ],
        "elasticsearch": [],
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
        ],
    },
)
